"""Statistics computation utils."""

import apache_beam  # noqa: F401
import numpy as np
import weatherbench2  # noqa: F401
import xarray as xr


def compute_statistics(dataset, vars, num_samples=100, single=False):
    """Compute statistics for single timestep.

    Args:
        dataset: xarray dataset.
        vars: list of features.
        num_samples (int, optional): _description_. Defaults to 100.
        single (bool, optional): if the features have multiple pressure levels. Defaults to False.

    Returns:
        means: dict with the means.
        stds: dict with the stds.
    """
    means = {}
    stds = {}
    for var in vars:
        print(f"Computing statistics for {var}")
        random_indexes = np.random.randint(0, len(dataset.time), num_samples)
        samples = data.isel(time=random_indexes)[var].values
        samples = np.nan_to_num(samples)
        axis_tuple = (0, 1, 2) if single else (0, 2, 3)
        means[var] = samples.mean(axis=axis_tuple)
        stds[var] = samples.std(axis=axis_tuple)
    return means, stds


def compute_statistics_diff(dataset, vars, num_samples=100, single=False, timestep=2):
    """Compute statistics for difference of two timesteps.

    Args:
        dataset: xarray dataset.
        vars: list of features.
        num_samples (int, optional): _description_. Defaults to 100.
        single (bool, optional): if the features have multiple pressure levels. Defaults to False.
        timestep (int, optional): number of steps to consider between start and end. Defaults to 2.

    Returns:
        means: dict with the means.
        stds: dict with the stds.
    """
    means = {}
    stds = {}
    for var in vars:
        print(f"Computing statistics for {var}")
        random_indexes = np.random.randint(0, len(dataset.time), num_samples)
        samples_start = data.isel(time=random_indexes)[var].values
        samples_start = np.nan_to_num(samples_start)
        samples_end = data.isel(time=random_indexes + timestep)[var].values
        samples_end = np.nan_to_num(samples_end)
        axis_tuple = (0, 1, 2) if single else (0, 2, 3)
        means[var] = (samples_end - samples_start).mean(axis=axis_tuple)
        stds[var] = (samples_end - samples_start).std(axis=axis_tuple)
    return means, stds


atmospheric_features = [
    "geopotential",
    "specific_humidity",
    "temperature",
    "u_component_of_wind",
    "v_component_of_wind",
    "vertical_velocity",
]

single_features = [
    "2m_temperature",
    "10m_u_component_of_wind",
    "10m_v_component_of_wind",
    "mean_sea_level_pressure",
    # "sea_surface_temperature",
    "total_precipitation_12hr",
]

static_features = ["geopotential_at_surface", "land_sea_mask"]

obs_path = "gs://weatherbench2/datasets/era5/1959-2022-6h-64x32_equiangular_conservative.zarr"
# obs_path = 'gs://weatherbench2/datasets/era5/1959-2022-6h-1440x721.zarr'
# obs_path = 'gs://weatherbench2/datasets/era5/1959-2022-6h-512x256_equiangular_conservative.zarr'
data = xr.open_zarr(obs_path)
num_samples = 100
means, stds = compute_statistics_diff(data, single_features, num_samples=num_samples, single=True)
print("Means: ", means)
print("Stds: ", stds)



# Copyright 2023 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Source: https://github.com/google-deepmind/graphcast.
"""Tools for converting from regular grids on a sphere, to triangular meshes."""

import numpy as np
import scipy
import trimesh

from graph_weather.models.gencast.graph import icosahedral_mesh


def _grid_lat_lon_to_coordinates(
    grid_latitude: np.ndarray, grid_longitude: np.ndarray
) -> np.ndarray:
    """Lat [num_lat] lon [num_lon] to 3d coordinates [num_lat, num_lon, 3]."""
    # Convert to spherical coordinates phi and theta defined in the grid.
    # Each [num_latitude_points, num_longitude_points]
    phi_grid, theta_grid = np.meshgrid(np.deg2rad(grid_longitude), np.deg2rad(90 - grid_latitude))

    # [num_latitude_points, num_longitude_points, 3]
    # Note this assumes unit radius, since for now we model the earth as a
    # sphere of unit radius, and keep any vertical dimension as a regular grid.
    return np.stack(
        [
            np.cos(phi_grid) * np.sin(theta_grid),
            np.sin(phi_grid) * np.sin(theta_grid),
            np.cos(theta_grid),
        ],
        axis=-1,
    )


def radius_query_indices(
    *,
    grid_latitude: np.ndarray,
    grid_longitude: np.ndarray,
    mesh: icosahedral_mesh.TriangularMesh,
    radius: float
) -> tuple[np.ndarray, np.ndarray]:
    """Returns mesh-grid edge indices for radius query.

    Args:
      grid_latitude: Latitude values for the grid [num_lat_points]
      grid_longitude: Longitude values for the grid [num_lon_points]
      mesh: Mesh object.
      radius: Radius of connectivity in R3. for a sphere of unit radius.

    Returns:
      tuple with `grid_indices` and `mesh_indices` indicating edges between the
      grid and the mesh such that the distances in a straight line (not geodesic)
      are smaller than or equal to `radius`.
      * grid_indices: Indices of shape [num_edges], that index into a
        [num_lat_points, num_lon_points] grid, after flattening the leading axes.
      * mesh_indices: Indices of shape [num_edges], that index into mesh.vertices.
    """

    # [num_grid_points=num_lat_points * num_lon_points, 3]
    grid_positions = _grid_lat_lon_to_coordinates(grid_latitude, grid_longitude).reshape([-1, 3])

    # [num_mesh_points, 3]
    mesh_positions = mesh.vertices
    kd_tree = scipy.spatial.cKDTree(mesh_positions)

    # [num_grid_points, num_mesh_points_per_grid_point]
    # Note `num_mesh_points_per_grid_point` is not constant, so this is a list
    # of arrays, rather than a 2d array.
    query_indices = kd_tree.query_ball_point(x=grid_positions, r=radius)

    grid_edge_indices = []
    mesh_edge_indices = []
    for grid_index, mesh_neighbors in enumerate(query_indices):
        grid_edge_indices.append(np.repeat(grid_index, len(mesh_neighbors)))
        mesh_edge_indices.append(mesh_neighbors)

    # [num_edges]
    grid_edge_indices = np.concatenate(grid_edge_indices, axis=0).astype(int)
    mesh_edge_indices = np.concatenate(mesh_edge_indices, axis=0).astype(int)

    return grid_edge_indices, mesh_edge_indices


def in_mesh_triangle_indices(
    *, grid_latitude: np.ndarray, grid_longitude: np.ndarray, mesh: icosahedral_mesh.TriangularMesh
) -> tuple[np.ndarray, np.ndarray]:
    """Returns mesh-grid edge indices for grid points contained in mesh triangles.

    Args:
      grid_latitude: Latitude values for the grid [num_lat_points]
      grid_longitude: Longitude values for the grid [num_lon_points]
      mesh: Mesh object.

    Returns:
      tuple with `grid_indices` and `mesh_indices` indicating edges between the
      grid and the mesh vertices of the triangle that contain each grid point.
      The number of edges is always num_lat_points * num_lon_points * 3
      * grid_indices: Indices of shape [num_edges], that index into a
        [num_lat_points, num_lon_points] grid, after flattening the leading axes.
      * mesh_indices: Indices of shape [num_edges], that index into mesh.vertices.
    """

    # [num_grid_points=num_lat_points * num_lon_points, 3]
    grid_positions = _grid_lat_lon_to_coordinates(grid_latitude, grid_longitude).reshape([-1, 3])

    mesh_trimesh = trimesh.Trimesh(vertices=mesh.vertices, faces=mesh.faces)

    # [num_grid_points] with mesh face indices for each grid point.
    _, _, query_face_indices = trimesh.proximity.closest_point(mesh_trimesh, grid_positions)

    # [num_grid_points, 3] with mesh node indices for each grid point.
    mesh_edge_indices = mesh.faces[query_face_indices]

    # [num_grid_points, 3] with grid node indices, where every row simply contains
    # the row (grid_point) index.
    grid_indices = np.arange(grid_positions.shape[0])
    grid_edge_indices = np.tile(grid_indices.reshape([-1, 1]), [1, 3])

    # Flatten to get a regular list.
    # [num_edges=num_grid_points*3]
    mesh_edge_indices = mesh_edge_indices.reshape([-1])
    grid_edge_indices = grid_edge_indices.reshape([-1])

    return grid_edge_indices, mesh_edge_indices



"""The weighted loss function for GenCast training."""

from typing import Optional

import numpy as np
import torch


class WeightedMSELoss(torch.nn.Module):
    """Module WeightedMSELoss.

    This module implement the loss described in GenCast's paper.
    """

    def __init__(
        self,
        grid_lat: Optional[torch.Tensor] = None,
        pressure_levels: Optional[torch.Tensor] = None,
        num_atmospheric_features: Optional[int] = None,
        single_features_weights: Optional[torch.Tensor] = None,
    ):
        """Initialize the WeightedMSELoss Module.

        More details about the features weights are reported in GraphCast's paper. In short, if the
        single features are "2m_temperature", "10m_u_component_of_wind", "10m_v_component_of_wind",
        "mean_sea_level_pressure" and "total_precipitation_12hr", then it's suggested to set
        corresponding weights as 1, 0.1, 0.1, 0.1 and 0.1.

        Args:
            grid_lat (torch.Tensor, optional): 1D tensor containing all the latitudes.
            pressure_levels (torch.Tensor, optional): 1D tensor containing all the pressure levels
                per variable.
            num_atmospheric_features (int, optional): number of atmospheric features.
            single_features_weights (torch.Tensor, optional): 1D tensor containing single features
                weights.
        """
        super().__init__()

        self.area_weights = None
        self.features_weights = None

        if grid_lat is not None:
            self.area_weights = torch.cos(grid_lat * np.pi / 180.0)

        if (
            pressure_levels is not None
            and num_atmospheric_features is not None
            and single_features_weights is not None
        ):
            pressure_weights = pressure_levels / torch.sum(pressure_levels)
            self.features_weights = torch.cat(
                (pressure_weights.repeat(num_atmospheric_features), single_features_weights), dim=-1
            )
        elif (
            pressure_levels is not None
            or num_atmospheric_features is not None
            or single_features_weights is not None
        ):
            raise ValueError(
                "Please to use features weights provide all three: pressure_levels,"
                "num_atmospheric_features and single_features_weights."
            )

        self.sigma_data = 1  # assuming normalized data!

    def _lambda_sigma(self, noise_level):
        noise_weights = (noise_level**2 + self.sigma_data**2) / (noise_level * self.sigma_data) ** 2
        return noise_weights  # [batch, 1]

    def forward(
        self,
        pred: torch.Tensor,
        noise_level: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        """Compute the loss.

        Args:
            pred (torch.Tensor): prediction of the model [batch, lon, lat, var].
            noise_level (torch.Tensor): noise levels fed to the model for the corresponding
                predictions [batch, 1].
            target (torch.Tensor): target tensor [batch, lon, lat, var].

        Returns:
            torch.Tensor: weighted MSE loss.
        """
        # check shapes
        if not (pred.shape == target.shape):
            raise ValueError(
                "Predictions and targets must have same shape. The actual shapes "
                f"are {pred.shape} and {target.shape}."
            )
        if not (len(pred.shape) == 4):
            raise ValueError(
                "The expected shape for predictions and targets is "
                f"[batch, lon, lat, var], but got {pred.shape}."
            )
        if not (noise_level.shape == (pred.shape[0], 1)):
            raise ValueError(
                f"The expected shape for noise levels is [batch, 1], but got {noise_level.shape}."
            )

        # compute square residuals
        loss = (pred - target) ** 2  # [batch, lon, lat, var]
        if torch.isnan(loss).any():
            raise ValueError("NaN values encountered in loss calculation.")

        # apply area and features weights to residuals
        if self.area_weights is not None:
            if not (len(self.area_weights) == pred.shape[2]):
                raise ValueError(
                    f"The size of grid_lat at initialization ({len(self.area_weights)}) "
                    f"and the number of latitudes in predictions ({pred.shape[2]}) "
                    "don't match."
                )
            loss *= self.area_weights[None, None, :, None]

        if self.features_weights is not None:
            if not (len(self.features_weights) == pred.shape[-1]):
                raise ValueError(
                    f"The size of features weights at initialization ({len(self.features_weights)})"
                    f" and the number of features in predictions ({pred.shape[-1]}) "
                    "don't match."
                )
            loss *= self.features_weights[None, None, None, :]

        # compute means across lon, lat, var for each sample in the batch
        loss = loss.flatten(1).mean(-1)  # [batch]

        # weight each sample using the corresponding noise level, then return the mean.
        loss *= self._lambda_sigma(noise_level).flatten()
        return loss.mean()



import h3
import numpy as np
import torch

from graph_weather import GraphWeatherAssimilator, GraphWeatherForecaster
from graph_weather.models import (
    AssimilatorDecoder,
    AssimilatorEncoder,
    Decoder,
    Encoder,
    Processor,
    ImageMetaModel,
    MetaModel,
    WrapperImageModel,
    WrapperMetaModel,
)
from graph_weather.models.losses import NormalizedMSELoss

from graph_weather.models.gencast.utils.noise import (
    generate_isotropic_noise,
    sample_noise_level,
)
from graph_weather.models.gencast import GraphBuilder, WeightedMSELoss, Denoiser


def test_encoder():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = Encoder(lat_lons).eval()

    features = torch.randn((2, len(lat_lons), 78))
    with torch.no_grad():
        x, edge_idx, edge_attr = model(features)
    assert x.size() == (5882 * 2, 256)
    assert edge_idx.size() == (2, 41162 * 2)


def test_encoder_uneven_grid():
    lat_lons = []
    for lat in range(-90, 90, 7):
        for lon in range(0, 180, 5):
            lat_lons.append((lat, lon))
        for lon in range(180, 360, 9):
            lat_lons.append((lat, lon))
    model = Encoder(lat_lons).eval()

    features = torch.randn((2, len(lat_lons), 78))
    with torch.no_grad():
        x, edge_idx, edge_attr = model(features)
    assert x.size() == (5882 * 2, 256)
    assert edge_idx.size() == (2, 41162 * 2)


def test_assimilation_encoder_uneven_grid():
    lat_lons = []
    for lat in range(-90, 90, 7):
        for lon in range(0, 180, 5):
            lat_lons.append((lat, lon, np.random.random(1)))
        for lon in range(180, 360, 9):
            lat_lons.append((lat, lon, np.random.random(1)))
    model = AssimilatorEncoder().eval()

    features = torch.randn((2, len(lat_lons), 2))
    with torch.no_grad():
        x, edge_idx, edge_attr = model(features, torch.tensor(lat_lons))
    assert x.size() == (5882 * 2, 256)
    assert edge_idx.size() == (2, 41162 * 2)


def test_processor():
    processor = Processor().eval()
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = Encoder(lat_lons).eval()

    features = torch.randn((3, len(lat_lons), 78))
    with torch.no_grad():
        x, edge_idx, edge_attr = model(features)
        out = processor(x, edge_idx, edge_attr)
    assert out.size() == x.size()


def test_decoder():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = Decoder(lat_lons).eval()
    features = torch.randn((3, len(lat_lons), 78))
    processed = torch.randn((3 * h3.num_hexagons(2), 256))
    with torch.no_grad():
        x = model(processed, features)
    assert x.size() == (3, 2592, 78)


def test_assimilator():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = AssimilatorDecoder(lat_lons).eval()
    processed = torch.randn((3 * h3.num_hexagons(2), 256))
    with torch.no_grad():
        x = model(processed, 3)
    assert x.size() == (3, 2592, 78)


def test_end2end():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = Encoder(lat_lons).eval()
    processor = Processor().eval()
    decoder = Decoder(lat_lons).eval()
    features = torch.randn((4, len(lat_lons), 78))
    with torch.no_grad():
        x, edge_idx, edge_attr = model(features)
        out = processor(x, edge_idx, edge_attr)
        pred = decoder(out, features)
    assert pred.size() == (4, 2592, 78)


def test_forecaster():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    model = GraphWeatherForecaster(lat_lons)
    # Add in auxiliary features
    features = torch.randn((1, len(lat_lons), 78 + 24))

    out = model(features)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()


def test_assimilator_model():
    obs_lat_lons = []
    for lat in range(-90, 90, 7):
        for lon in range(0, 180, 6):
            obs_lat_lons.append((lat, lon, np.random.random(1)))
        for lon in 360 * np.random.random(100):
            obs_lat_lons.append((lat, lon, np.random.random(1)))

    output_lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            output_lat_lons.append((lat, lon))
    model = GraphWeatherAssimilator(output_lat_lons=output_lat_lons, analysis_dim=24)

    features = torch.randn((1, len(obs_lat_lons), 2))
    lat_lon_heights = torch.tensor(obs_lat_lons)
    out = model(features, lat_lon_heights)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()


def test_forecaster_and_loss():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    criterion = NormalizedMSELoss(lat_lons=lat_lons, feature_variance=torch.randn((78,)))
    model = GraphWeatherForecaster(lat_lons)
    # Add in auxiliary features
    features = torch.randn((2, len(lat_lons), 78 + 24))

    out = model(features)
    loss = criterion(out, torch.rand(out.shape))
    assert not torch.isnan(loss)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()
    loss.backward()


def test_assimilator_model_grad_checkpoint():
    obs_lat_lons = []
    for lat in range(-90, 90, 7):
        for lon in range(0, 180, 6):
            obs_lat_lons.append((lat, lon, np.random.random(1)))
        for lon in 360 * np.random.random(100):
            obs_lat_lons.append((lat, lon, np.random.random(1)))

    output_lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            output_lat_lons.append((lat, lon))
    model = GraphWeatherAssimilator(
        output_lat_lons=output_lat_lons, analysis_dim=24, use_checkpointing=True
    )

    features = torch.randn((1, len(obs_lat_lons), 2))
    lat_lon_heights = torch.tensor(obs_lat_lons)
    out = model(features, lat_lon_heights)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()


def test_forecaster_and_loss_grad_checkpoint():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))
    criterion = NormalizedMSELoss(lat_lons=lat_lons, feature_variance=torch.randn((78,)))
    model = GraphWeatherForecaster(lat_lons, use_checkpointing=True)
    # Add in auxiliary features
    features = torch.randn((2, len(lat_lons), 78 + 24))

    out = model(features)
    loss = criterion(out, torch.rand(out.shape))
    assert not torch.isnan(loss)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()
    loss.backward()


def test_normalized_loss():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))

    # Generate output as strictly positive random features
    out = torch.rand((2, len(lat_lons), 78)) + 0.0001
    feature_variance = out**2
    target = torch.zeros((2, len(lat_lons), 78))

    criterion = NormalizedMSELoss(
        lat_lons=lat_lons, feature_variance=feature_variance, normalize=True
    )
    loss = criterion(out, target)

    assert not torch.isnan(loss)
    # Since feature_variance = out**2 and target = 0, we expect loss = weights
    assert torch.isclose(loss, criterion.weights.expand_as(out.mean(-1)).mean())


def test_image_meta_model():
    batch = 2
    channels = 3
    size = 4
    patch_size = 2
    image = torch.randn((batch, channels, size, size))
    model = ImageMetaModel(
        image_size=size,
        patch_size=patch_size,
        channels=channels,
        depth=1,
        heads=1,
        mlp_dim=7,
        dim_head=64,
    )

    out = model(image)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()
    assert out.size() == image.size()


def test_wrapper_image_meta_model():
    batch = 2
    channels = 3
    size = 4
    patch_size = 2
    model = ImageMetaModel(
        image_size=size,
        patch_size=patch_size,
        channels=channels,
        depth=1,
        heads=1,
        mlp_dim=7,
        dim_head=64,
    )
    scale_factor = 3
    big_image = torch.randn((batch, channels, size * scale_factor, size * scale_factor))
    big_model = WrapperImageModel(model, scale_factor)
    out = big_model(big_image)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()
    assert out.size() == big_image.size()


def test_meta_model():
    lat_lons = []
    for lat in range(-90, 90, 5):
        for lon in range(0, 360, 5):
            lat_lons.append((lat, lon))

    batch = 2
    channels = 3
    image_size = 20
    patch_size = 4
    model = MetaModel(
        lat_lons,
        image_size=image_size,
        patch_size=patch_size,
        depth=1,
        heads=1,
        mlp_dim=7,
        channels=channels,
        dim_head=64,
    )
    features = torch.randn((batch, len(lat_lons), channels))

    out = model(features)
    assert not torch.isnan(out).any()
    assert not torch.isnan(out).any()
    assert out.size() == features.size()


def test_gencast_noise():
    num_lat = 32
    num_samples = 5
    target_residuals = np.zeros((2 * num_lat, num_lat, num_samples))
    noise_level = sample_noise_level()
    noise = generate_isotropic_noise(num_lat=num_lat, num_samples=target_residuals.shape[-1])
    corrupted_residuals = target_residuals + noise_level * noise
    assert corrupted_residuals.shape == target_residuals.shape
    assert not np.isnan(corrupted_residuals).any()


def test_gencast_graph():
    grid_lat = np.arange(-90, 90, 1)
    grid_lon = np.arange(0, 360, 1)
    graphs = GraphBuilder(grid_lon=grid_lon, grid_lat=grid_lat, splits=0, num_hops=1)

    assert graphs.mesh_graph.x.shape[0] == 12
    assert graphs.g2m_graph["grid_nodes"].x.shape[0] == 360 * 180
    assert graphs.m2g_graph["mesh_nodes"].x.shape[0] == 12
    assert not torch.isnan(graphs.mesh_graph.edge_attr).any()
    assert graphs.khop_mesh_graph.x.shape[0] == 12
    assert graphs.khop_mesh_graph.edge_attr.shape[0] == 12 * 10


def test_gencast_loss():
    grid_lat = torch.arange(-90, 90, 1)
    grid_lon = torch.arange(0, 360, 1)
    pressure_levels = torch.tensor(
        [50.0, 100.0, 150.0, 200.0, 250, 300, 400, 500, 600, 700, 850, 925, 1000.0]
    )
    single_features_weights = torch.tensor([1, 0.1, 0.1, 0.1, 0.1])
    num_atmospheric_features = 6
    batch_size = 3
    features_dim = len(pressure_levels) * num_atmospheric_features + len(single_features_weights)

    loss = WeightedMSELoss(
        grid_lat=grid_lat,
        pressure_levels=pressure_levels,
        num_atmospheric_features=num_atmospheric_features,
        single_features_weights=single_features_weights,
    )

    preds = torch.rand((batch_size, len(grid_lon), len(grid_lat), features_dim))
    noise_levels = torch.rand((batch_size, 1))
    targets = torch.rand((batch_size, len(grid_lon), len(grid_lat), features_dim))
    assert loss.forward(preds, noise_levels, targets) is not None


def test_gencast_denoiser():
    grid_lat = np.arange(-90, 90, 1)
    grid_lon = np.arange(0, 360, 1)
    input_features_dim = 10
    output_features_dim = 5
    batch_size = 3

    denoiser = Denoiser(
        grid_lon=grid_lon,
        grid_lat=grid_lat,
        input_features_dim=input_features_dim,
        output_features_dim=output_features_dim,
        hidden_dims=[16, 32],
        num_blocks=3,
        num_heads=4,
        splits=0,
        num_hops=1,
        device=torch.device("cpu"),
    ).eval()

    corrupted_targets = torch.randn((batch_size, len(grid_lon), len(grid_lat), output_features_dim))
    prev_inputs = torch.randn((batch_size, len(grid_lon), len(grid_lat), 2 * input_features_dim))
    noise_levels = torch.randn((batch_size, 1))

    with torch.no_grad():
        preds = denoiser(
            corrupted_targets=corrupted_targets, prev_inputs=prev_inputs, noise_levels=noise_levels
        )

    assert not torch.isnan(preds).any()



"""Model for forecasting weather from NWP states"""

import torch
from huggingface_hub import PyTorchModelHubMixin

from graph_weather.models import AssimilatorDecoder, AssimilatorEncoder, Processor


class GraphWeatherAssimilator(torch.nn.Module, PyTorchModelHubMixin):
    """Model to generate analysis file from raw observations"""

    def __init__(
        self,
        output_lat_lons: list,
        resolution: int = 2,
        observation_dim: int = 2,
        analysis_dim: int = 78,
        node_dim: int = 256,
        edge_dim: int = 256,
        num_blocks: int = 9,
        hidden_dim_processor_node: int = 256,
        hidden_dim_processor_edge: int = 256,
        hidden_layers_processor_node: int = 2,
        hidden_layers_processor_edge: int = 2,
        hidden_dim_decoder: int = 128,
        hidden_layers_decoder: int = 2,
        norm_type: str = "LayerNorm",
        use_checkpointing: bool = False,
    ):
        """
        Graph Weather Data Assimilation model

        Args:
        observation_lat_lons: Lat/lon points of the observations
        output_lat_lons: List of latitude and longitudes for the output analysis
        resolution: Resolution of the H3 grid, prefer even resolutions, as
            odd ones have octogons and heptagons as well
        observation_dim: Input feature size
        analysis_dim: Output Analysis feature dim
        node_dim: Node hidden dimension
        edge_dim: Edge hidden dimension
        num_blocks: Number of message passing blocks in the Processor
        hidden_dim_processor_node: Hidden dimension of the node processors
        hidden_dim_processor_edge: Hidden dimension of the edge processors
        hidden_layers_processor_node: Number of hidden layers in the node processors
        hidden_layers_processor_edge: Number of hidden layers in the edge processors
        hidden_dim_decoder:Number of hidden dimensions in the decoder
        hidden_layers_decoder: Number of layers in the decoder
        norm_type: Type of norm for the MLPs
            one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
        use_checkpointing: Whether to use gradient checkpointing or not
        """
        super().__init__()

        self.encoder = AssimilatorEncoder(
            resolution=resolution,
            input_dim=observation_dim,
            output_dim=node_dim,
            output_edge_dim=edge_dim,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
            use_checkpointing=use_checkpointing,
        )
        self.processor = Processor(
            input_dim=node_dim,
            edge_dim=edge_dim,
            num_blocks=num_blocks,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
        )
        self.decoder = AssimilatorDecoder(
            lat_lons=output_lat_lons,
            resolution=resolution,
            input_dim=node_dim,
            output_dim=analysis_dim,
            output_edge_dim=edge_dim,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
            hidden_dim_decoder=hidden_dim_decoder,
            hidden_layers_decoder=hidden_layers_decoder,
            use_checkpointing=use_checkpointing,
        )

    def forward(self, features: torch.Tensor, obs_lat_lon_heights: torch.Tensor) -> torch.Tensor:
        """
        Compute the analysis output

        Args:
            features: The input features, aligned with the order of lat_lons_heights
            obs_lat_lon_heights: Observation lat/lon/heights in same order as features

        Returns:
            The next state in the forecast
        """
        x, edge_idx, edge_attr = self.encoder(features, obs_lat_lon_heights)
        x = self.processor(x, edge_idx, edge_attr)
        x = self.decoder(x, features.shape[0])
        return x



"""Decoders to decode from the Processor graph to the original graph with updated values

In the original paper the decoder is described as

The Decoder maps back to physical data defined on a latitude/longitude grid. The underlying graph is
again bipartite, this time mapping icosahedron→lat/lon.
The inputs to the Decoder come from the Processor, plus a skip connection back to the original
state of the 78 atmospheric variables onthe latitude/longitude grid.
The output of the Decoder is the predicted 6-hour change in the 78 atmospheric variables,
which is then added to the initial state to produce the new state. We found 6 hours to be a good
balance between shorter time steps (simpler dynamics to model but more iterations required during
rollout) and longer time steps (fewer iterations required during rollout but modeling
more complex dynamics)

"""

import torch

from graph_weather.models.layers.assimilator_decoder import AssimilatorDecoder


class Decoder(AssimilatorDecoder):
    """Decoder graph module"""

    def __init__(
        self,
        lat_lons,
        resolution: int = 2,
        input_dim: int = 256,
        output_dim: int = 78,
        output_edge_dim: int = 256,
        hidden_dim_processor_node: int = 256,
        hidden_dim_processor_edge: int = 256,
        hidden_layers_processor_node: int = 2,
        hidden_layers_processor_edge: int = 2,
        mlp_norm_type: str = "LayerNorm",
        hidden_dim_decoder: int = 128,
        hidden_layers_decoder: int = 2,
        use_checkpointing: bool = False,
    ):
        """
        Decoder from latent graph to lat/lon graph

        Args:
            lat_lons: List of (lat,lon) points
            resolution: H3 resolution level
            input_dim: Input node dimension
            output_dim: Output node dimension
            output_edge_dim: Edge dimension
            hidden_dim_processor_node: Hidden dimension of the node processors
            hidden_dim_processor_edge: Hidden dimension of the edge processors
            hidden_layers_processor_node: Number of hidden layers in the node processors
            hidden_layers_processor_edge: Number of hidden layers in the edge processors
            hidden_dim_decoder:Number of hidden dimensions in the decoder
            hidden_layers_decoder: Number of layers in the decoder
            mlp_norm_type: Type of norm for the MLPs
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
            use_checkpointing: Whether to use gradient checkpointing or not
        """
        super().__init__(
            lat_lons,
            resolution,
            input_dim,
            output_dim,
            output_edge_dim,
            hidden_dim_processor_node,
            hidden_dim_processor_edge,
            hidden_layers_processor_node,
            hidden_layers_processor_edge,
            mlp_norm_type,
            hidden_dim_decoder,
            hidden_layers_decoder,
            use_checkpointing,
        )

    def forward(
        self, processor_features: torch.Tensor, start_features: torch.Tensor
    ) -> torch.Tensor:
        """
        Adds features to the encoding graph

        Args:
            processor_features: Processed features in shape [B*Nodes, Features]
            start_features: Original input features to the encoder, with shape [B, Nodes, Features]

        Returns:
            Updated features for model
        """
        out = super().forward(processor_features, start_features.shape[0])
        out = out + start_features  # residual connection
        return out



"""Modules"""

import torch
import torch.nn as nn
from torch_geometric.nn import MessagePassing


class MLP(nn.Module):
    """Classic multi-layer perceptron (MLP) module."""

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list[int],
        activation_layer: nn.Module = nn.ReLU,
        use_layer_norm: bool = False,
        bias: bool = True,
        activate_final: bool = False,
    ):
        """Initialize MLP module.

        Args:
            input_dim (int): dimension of input.
            hidden_dims (List[int]): list of hidden linear layers dimensions.
            activation_layer (torch.nn.Module): activation
                function to use. Defaults to torch.nn.ReLU.
            use_layer_norm (bool, optional): if Ttrue apply LayerNorm to output. Defaults to False.
            bias (bool, optional): if true use bias in linear layers. Defaults to True.
            activate_final (bool, optional): whether to apply  the activation function to the final
                layer. Defaults to False.
        """
        super().__init__()

        # Initialize linear layers
        self.linears = nn.ModuleList()
        self.linears.append(nn.Linear(input_dim, hidden_dims[0], bias=bias))
        for i in range(0, len(hidden_dims) - 1):
            self.linears.append(nn.Linear(hidden_dims[i], hidden_dims[i + 1], bias=bias))

        # Initialize activation
        self.activation = activation_layer()

        # Initialize layer normalization
        self.norm_layer = None
        if use_layer_norm:
            self.norm_layer = nn.LayerNorm(hidden_dims[-1])

        self.activate_final = activate_final

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Apply MLP to input."""

        for linear in self.linears[:-1]:
            x = linear(x)
            x = self.activation(x)

        x = self.linears[-1](x)

        if self.activate_final:
            x = self.activation(x)

        if self.norm_layer is not None:
            x = self.norm_layer(x)
        return x


class InteractionNetwork(MessagePassing):
    """Single message-passing interaction network as described in GenCast.

    This network performs two steps:
    1) message-passing: e'_ij = MLP([e_ij,v_i,v_j])
    2) aggregation: v'_j = MLP([v_j, sum_i {e'_ij}])
    The underlying graph is a directed graph.

    Note:
        We don't need to update edges in GenCast, hence we skip it.
    """

    def __init__(
        self,
        sender_dim: int,
        receiver_dim: int,
        edge_attr_dim: int,
        hidden_dims: list[int],
        use_layer_norm: bool = False,
        activation_layer: nn.Module = nn.ReLU,
    ):
        """Initialize the Interaction Network.

        Args:
            sender_dim (int): dimension of sender nodes' features.
            receiver_dim (int): dimension of receiver nodes' features.
            edge_attr_dim (int): dimension of the edge features.
            hidden_dims (list[int]): list of sizes of MLP's linear layers.
            use_layer_norm (bool): if true add layer normalization to MLP's last layer.
                Defaults to False.
            activation_layer (torch.nn.Module): activation function. Defaults to nn.ReLU.
        """
        super().__init__(aggr="add", flow="source_to_target")
        self.mlp_edges = MLP(
            input_dim=sender_dim + receiver_dim + edge_attr_dim,
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )
        self.mlp_nodes = MLP(
            input_dim=receiver_dim + hidden_dims[-1],
            hidden_dims=hidden_dims,
            activation_layer=activation_layer,
            use_layer_norm=use_layer_norm,
            bias=True,
            activate_final=False,
        )

    def message(self, x_i, x_j, edge_attr):
        """Message-passing step."""
        x = torch.cat((x_i, x_j, edge_attr), dim=-1)
        x = self.mlp_edges(x)
        return x

    def forward(
        self,
        x: tuple[torch.Tensor, torch.Tensor],
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the output of the Interaction Network.

        This method processes the input node features and edge attributes through
        the network to produce the output node features.

        Args:
            x (tuple[torch.Tensor, torch.Tensor]): a tuple containing:
                - sender nodes' features (torch.Tensor): features of the sender nodes.
                - receiver nodes' features (torch.Tensor): features of the receiver nodes.
            edge_index (torch.Tensor): tensor containing edge indices, defining the
                connections between nodes.
            edge_attr (torch.Tensor): tensor containing edge features, representing
                the attributes of each edge.

        Returns:
            torch.Tensor: the resulting node features after applying the Interaction Network.
        """
        aggr = self.propagate(
            edge_index, x=x, edge_attr=edge_attr, size=(x[0].shape[0], x[1].shape[0])
        )
        out = self.mlp_nodes(torch.cat((x[1], aggr), dim=-1))
        return out



"""Processor for the latent graph

In the original paper the processor is described as

The Processor iteratively processes the 256-channel latent feature data on the icosahedron grid
using 9 rounds of message-passing GNNs. During each round, a node exchanges information with itself
and its immediate neighbors. There are residual connections between each round of processing.

"""

import torch

from graph_weather.models.layers.graph_net_block import GraphProcessor


class Processor(torch.nn.Module):
    """Processor for latent graphD"""

    def __init__(
        self,
        input_dim: int = 256,
        edge_dim: int = 256,
        num_blocks: int = 9,
        hidden_dim_processor_node: int = 256,
        hidden_dim_processor_edge: int = 256,
        hidden_layers_processor_node: int = 2,
        hidden_layers_processor_edge: int = 2,
        mlp_norm_type: str = "LayerNorm",
    ):
        """
        Latent graph processor

        Args:
            input_dim: Input dimension for the node
            edge_dim: Edge input dimension
            num_blocks: Number of message passing blocks
            hidden_dim_processor_node: Hidden dimension of the node processors
            hidden_dim_processor_edge: Hidden dimension of the edge processors
            hidden_layers_processor_node: Number of hidden layers in the node processors
            hidden_layers_processor_edge: Number of hidden layers in the edge processors
            mlp_norm_type: Type of norm for the MLPs
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
        """
        super().__init__()
        # Build the default graph
        # Take features from encoder and put into processor graph
        self.input_dim = input_dim

        self.graph_processor = GraphProcessor(
            num_blocks,
            input_dim,
            edge_dim,
            hidden_dim_processor_node,
            hidden_dim_processor_edge,
            hidden_layers_processor_node,
            hidden_layers_processor_edge,
            mlp_norm_type,
        )

    def forward(self, x: torch.Tensor, edge_index, edge_attr) -> torch.Tensor:
        """
        Adds features to the encoding graph

        Args:
            x: Torch tensor containing node features
            edge_index: Connectivity of graph, of shape [2, Num edges] in COO format
            edge_attr: Edge attribues in [Num edges, Features] shape

        Returns:
            torch Tensor containing the values of the nodes of the graph
        """
        out, _ = self.graph_processor(x, edge_index, edge_attr)
        return out



"""Weather loss functions"""

import numpy as np
import torch


class NormalizedMSELoss(torch.nn.Module):
    """Loss function described in the paper"""

    def __init__(
        self, feature_variance: list, lat_lons: list, device="cpu", normalize: bool = False
    ):
        """
        Normalized MSE Loss as described in the paper

        This re-scales each physical variable such that it has unit-variance in the 3 hour temporal
        difference. E.g. for temperature data, divide every one at all pressure levels by
        sigma_t_3hr, where sigma^2_T,3hr is the variance of the 3 hour change in temperature,
         averaged across space (lat/lon + pressure levels) and time (100 random temporal frames).

         Additionally weights by the cos(lat) of the feature

         cos and sin should be in radians

        Args:
            feature_variance: Variance for each of the physical features
            lat_lons: List of lat/lon pairs, used to generate weighting
            device: checks for device whether it supports gpu or not
            normalize: option for normalize
        """
        # TODO Rescale by nominal static air density at each pressure level
        super().__init__()
        self.feature_variance = torch.tensor(feature_variance)
        assert not torch.isnan(self.feature_variance).any()
        weights = []
        for lat, lon in lat_lons:
            weights.append(np.cos(lat * np.pi / 180.0))
        self.weights = torch.tensor(weights, dtype=torch.float)
        self.normalize = normalize
        assert not torch.isnan(self.weights).any()

    def forward(self, pred: torch.Tensor, target: torch.Tensor):
        """
        Calculate the loss

        Rescales both predictions and target, so assumes neither are already normalized
        Additionally weights by the cos(lat) of the set of features

        Args:
            pred: Prediction tensor
            target: Target tensor

        Returns:
            MSE loss on the variance-normalized values
        """
        self.feature_variance = self.feature_variance.to(pred.device)
        self.weights = self.weights.to(pred.device)

        out = (pred - target) ** 2

        if self.normalize:
            out = out / self.feature_variance

        assert not torch.isnan(out).any()
        # Mean of the physical variables
        out = out.mean(-1)
        # Weight by the latitude, as that changes, so does the size of the pixel
        out = out * self.weights.expand_as(out)
        assert not torch.isnan(out).any()
        return out.mean()



"""Model for forecasting weather from NWP states"""

from typing import Optional

import torch
from huggingface_hub import PyTorchModelHubMixin

from graph_weather.models import Decoder, Encoder, Processor


class GraphWeatherForecaster(torch.nn.Module, PyTorchModelHubMixin):
    """Main weather prediction model from the paper"""

    def __init__(
        self,
        lat_lons: list,
        resolution: int = 2,
        feature_dim: int = 78,
        aux_dim: int = 24,
        output_dim: Optional[int] = None,
        node_dim: int = 256,
        edge_dim: int = 256,
        num_blocks: int = 9,
        hidden_dim_processor_node: int = 256,
        hidden_dim_processor_edge: int = 256,
        hidden_layers_processor_node: int = 2,
        hidden_layers_processor_edge: int = 2,
        hidden_dim_decoder: int = 128,
        hidden_layers_decoder: int = 2,
        norm_type: str = "LayerNorm",
        use_checkpointing: bool = False,
    ):
        """
        Graph Weather Model based off https://arxiv.org/pdf/2202.07575.pdf

        Args:
            lat_lons: List of latitude and longitudes for the grid
            resolution: Resolution of the H3 grid, prefer even resolutions, as
                odd ones have octogons and heptagons as well
            feature_dim: Input feature size
            aux_dim: Number of non-NWP features (i.e. landsea mask, lat/lon, etc)
            output_dim: Optional, output feature size, useful if want only subset of variables in
            output
            node_dim: Node hidden dimension
            edge_dim: Edge hidden dimension
            num_blocks: Number of message passing blocks in the Processor
            hidden_dim_processor_node: Hidden dimension of the node processors
            hidden_dim_processor_edge: Hidden dimension of the edge processors
            hidden_layers_processor_node: Number of hidden layers in the node processors
            hidden_layers_processor_edge: Number of hidden layers in the edge processors
            hidden_dim_decoder:Number of hidden dimensions in the decoder
            hidden_layers_decoder: Number of layers in the decoder
            norm_type: Type of norm for the MLPs
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
            use_checkpointing: Use gradient checkpointing to reduce model memory
        """
        super().__init__()
        self.feature_dim = feature_dim
        if output_dim is None:
            output_dim = self.feature_dim

        self.encoder = Encoder(
            lat_lons=lat_lons,
            resolution=resolution,
            input_dim=feature_dim + aux_dim,
            output_dim=node_dim,
            output_edge_dim=edge_dim,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
            use_checkpointing=use_checkpointing,
        )
        self.processor = Processor(
            input_dim=node_dim,
            edge_dim=edge_dim,
            num_blocks=num_blocks,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
        )
        self.decoder = Decoder(
            lat_lons=lat_lons,
            resolution=resolution,
            input_dim=node_dim,
            output_dim=output_dim,
            output_edge_dim=edge_dim,
            hidden_dim_processor_edge=hidden_dim_processor_edge,
            hidden_layers_processor_node=hidden_layers_processor_node,
            hidden_dim_processor_node=hidden_dim_processor_node,
            hidden_layers_processor_edge=hidden_layers_processor_edge,
            mlp_norm_type=norm_type,
            hidden_dim_decoder=hidden_dim_decoder,
            hidden_layers_decoder=hidden_layers_decoder,
            use_checkpointing=use_checkpointing,
        )

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """
        Compute the new state of the forecast

        Args:
            features: The input features, aligned with the order of lat_lons_heights

        Returns:
            The next state in the forecast
        """
        x, edge_idx, edge_attr = self.encoder(features)
        x = self.processor(x, edge_idx, edge_attr)
        x = self.decoder(x, features[..., : self.feature_dim])
        return x



"""Setup"""

from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
install_requires = (this_directory / "requirements.txt").read_text().splitlines()
long_description = (this_directory / "README.md").read_text()

setup(
    name="graph_weather",
    version="1.0.72",
    packages=find_packages(),
    url="https://github.com/openclimatefix/graph_weather",
    license="MIT License",
    company="Open Climate Fix Ltd",
    author="Jacob Bieker",
    install_requires=install_requires,
    extras_requires={"vis": ["matplotlib"], "data": ["pysolar"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="jacob@bieker.tech",
    description="Weather Forecasting with Graph Neural Networks",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
)



# Copyright 2023 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://github.com/google-deepmind/graphcast.
"""Utils for creating icosahedral meshes."""

from typing import List, NamedTuple, Tuple

import numpy as np
from scipy.spatial import transform


class TriangularMesh(NamedTuple):
    """Data structure for triangular meshes.

    Attributes:
      vertices: spatial positions of the vertices of the mesh of shape
          [num_vertices, num_dims].
      faces: triangular faces of the mesh of shape [num_faces, 3]. Contains
          integer indices into `vertices`.

    """

    vertices: np.ndarray
    faces: np.ndarray


def get_hierarchy_of_triangular_meshes_for_sphere(splits: int) -> List[TriangularMesh]:
    """Returns a sequence of meshes, each with triangularization sphere.

    Starting with a regular icosahedron (12 vertices, 20 faces, 30 edges) with
    circumscribed unit sphere. Then, each triangular face is iteratively
    subdivided into 4 triangular faces `splits` times. The new vertices are then
    projected back onto the unit sphere. All resulting meshes are returned in a
    list, from lowest to highest resolution.

    The vertices in each face are specified in counter-clockwise order as
    observed from the outside the icosahedron.

    Args:
       splits: How many times to split each triangle.

    Returns:
       Sequence of `TriangularMesh`s of length `splits + 1` each with:

         vertices: [num_vertices, 3] vertex positions in 3D, all with unit norm.
         faces: [num_faces, 3] with triangular faces joining sets of 3 vertices.
             Each row contains three indices into the vertices array, indicating
             the vertices adjacent to the face. Always with positive orientation
             (counterclock-wise when looking from the outside).
    """
    current_mesh = get_icosahedron()
    output_meshes = [current_mesh]
    for _ in range(splits):
        current_mesh = _two_split_unit_sphere_triangle_faces(current_mesh)
        output_meshes.append(current_mesh)
    return output_meshes


def get_icosahedron() -> TriangularMesh:
    """Returns a regular icosahedral mesh with circumscribed unit sphere.

    See https://en.wikipedia.org/wiki/Regular_icosahedron#Cartesian_coordinates
    for details on the construction of the regular icosahedron.

    The vertices in each face are specified in counter-clockwise order as observed
    from the outside of the icosahedron.

    Returns:
       TriangularMesh with:

       vertices: [num_vertices=12, 3] vertex positions in 3D, all with unit norm.
       faces: [num_faces=20, 3] with triangular faces joining sets of 3 vertices.
           Each row contains three indices into the vertices array, indicating
           the vertices adjacent to the face. Always with positive orientation (
           counterclock-wise when looking from the outside).

    """
    phi = (1 + np.sqrt(5)) / 2
    vertices = []
    for c1 in [1.0, -1.0]:
        for c2 in [phi, -phi]:
            vertices.append((c1, c2, 0.0))
            vertices.append((0.0, c1, c2))
            vertices.append((c2, 0.0, c1))

    vertices = np.array(vertices, dtype=np.float32)
    vertices /= np.linalg.norm([1.0, phi])

    # I did this manually, checking the orientation one by one.
    faces = [
        (0, 1, 2),
        (0, 6, 1),
        (8, 0, 2),
        (8, 4, 0),
        (3, 8, 2),
        (3, 2, 7),
        (7, 2, 1),
        (0, 4, 6),
        (4, 11, 6),
        (6, 11, 5),
        (1, 5, 7),
        (4, 10, 11),
        (4, 8, 10),
        (10, 8, 3),
        (10, 3, 9),
        (11, 10, 9),
        (11, 9, 5),
        (5, 9, 7),
        (9, 3, 7),
        (1, 6, 5),
    ]

    # By default the top is an aris parallel to the Y axis.
    # Need to rotate around the y axis by half the supplementary to the
    # angle between faces divided by two to get the desired orientation.
    #                          /O\  (top arist)
    #                     /          \                           Z
    # (adjacent face)/                    \  (adjacent face)     ^
    #           /     angle_between_faces      \                 |
    #      /                                        \            |
    #  /                                                 \      YO-----> X
    # This results in:
    #  (adjacent faceis now top plane)
    #  ----------------------O\  (top arist)
    #                           \
    #                             \
    #                               \     (adjacent face)
    #                                 \
    #                                   \
    #                                     \

    angle_between_faces = 2 * np.arcsin(phi / np.sqrt(3))
    rotation_angle = (np.pi - angle_between_faces) / 2
    rotation = transform.Rotation.from_euler(seq="y", angles=rotation_angle)
    rotation_matrix = rotation.as_matrix()
    vertices = np.dot(vertices, rotation_matrix)

    return TriangularMesh(
        vertices=vertices.astype(np.float32), faces=np.array(faces, dtype=np.int32)
    )


def _two_split_unit_sphere_triangle_faces(triangular_mesh: TriangularMesh) -> TriangularMesh:
    """Splits each triangular face into 4 triangles keeping the orientation."""

    # Every time we split a triangle into 4 we will be adding 3 extra vertices,
    # located at the edge centres.
    # This class handles the positioning of the new vertices, and avoids creating
    # duplicates.
    new_vertices_builder = _ChildVerticesBuilder(triangular_mesh.vertices)

    new_faces = []
    for ind1, ind2, ind3 in triangular_mesh.faces:
        # Transform each triangular face into 4 triangles,
        # preserving the orientation.
        #                    ind3
        #                   /    \
        #                /          \
        #              /      #3       \
        #            /                  \
        #         ind31 -------------- ind23
        #         /   \                /   \
        #       /       \     #4     /      \
        #     /    #1     \        /    #2    \
        #   /               \    /              \
        # ind1 ------------ ind12 ------------ ind2
        ind12 = new_vertices_builder.get_new_child_vertex_index((ind1, ind2))
        ind23 = new_vertices_builder.get_new_child_vertex_index((ind2, ind3))
        ind31 = new_vertices_builder.get_new_child_vertex_index((ind3, ind1))
        # Note how each of the 4 triangular new faces specifies the order of the
        # vertices to preserve the orientation of the original face. As the input
        # face should always be counter-clockwise as specified in the diagram,
        # this means child faces should also be counter-clockwise.
        new_faces.extend(
            [
                [ind1, ind12, ind31],  # 1
                [ind12, ind2, ind23],  # 2
                [ind31, ind23, ind3],  # 3
                [ind12, ind23, ind31],  # 4
            ]
        )
    return TriangularMesh(
        vertices=new_vertices_builder.get_all_vertices(), faces=np.array(new_faces, dtype=np.int32)
    )


class _ChildVerticesBuilder(object):
    """Bookkeeping of new child vertices added to an existing set of vertices."""

    def __init__(self, parent_vertices):
        # Because the same new vertex will be required when splitting adjacent
        # triangles (which share an edge) we keep them in a hash table indexed by
        # sorted indices of the vertices adjacent to the edge, to avoid creating
        # duplicated child vertices.
        self._child_vertices_index_mapping = {}
        self._parent_vertices = parent_vertices
        # We start with all previous vertices.
        self._all_vertices_list = list(parent_vertices)

    def _get_child_vertex_key(self, parent_vertex_indices):
        return tuple(sorted(parent_vertex_indices))

    def _create_child_vertex(self, parent_vertex_indices):
        """Creates a new vertex."""
        # Position for new vertex is the middle point, between the parent points,
        # projected to unit sphere.
        child_vertex_position = self._parent_vertices[list(parent_vertex_indices)].mean(0)
        child_vertex_position /= np.linalg.norm(child_vertex_position)

        # Add the vertex to the output list. The index for this new vertex will
        # match the length of the list before adding it.
        child_vertex_key = self._get_child_vertex_key(parent_vertex_indices)
        self._child_vertices_index_mapping[child_vertex_key] = len(self._all_vertices_list)
        self._all_vertices_list.append(child_vertex_position)

    def get_new_child_vertex_index(self, parent_vertex_indices):
        """Returns index for a child vertex, creating it if necessary."""
        # Get the key to see if we already have a new vertex in the middle.
        child_vertex_key = self._get_child_vertex_key(parent_vertex_indices)
        if child_vertex_key not in self._child_vertices_index_mapping:
            self._create_child_vertex(parent_vertex_indices)
        return self._child_vertices_index_mapping[child_vertex_key]

    def get_all_vertices(self):
        """Returns an array with old vertices."""
        return np.array(self._all_vertices_list)


def faces_to_edges(faces: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms polygonal faces to sender and receiver indices.

    It does so by transforming every face into N_i edges. Such if the triangular
    face has indices [0, 1, 2], three edges are added 0->1, 1->2, and 2->0.

    If all faces have consistent orientation, and the surface represented by the
    faces is closed, then every edge in a polygon with a certain orientation
    is also part of another polygon with the opposite orientation. In this
    situation, the edges returned by the method are always bidirectional.

    Args:
      faces: Integer array of shape [num_faces, 3]. Contains node indices
          adjacent to each face.

    Returns:
      Tuple with sender/receiver indices, each of shape [num_edges=num_faces*3].

    """
    assert faces.ndim == 2
    assert faces.shape[-1] == 3
    senders = np.concatenate([faces[:, 0], faces[:, 1], faces[:, 2]])
    receivers = np.concatenate([faces[:, 1], faces[:, 2], faces[:, 0]])
    return senders, receivers



"""Noise generation utils."""

import numpy as np
import pyshtools as pysh
import torch


def generate_isotropic_noise(num_lat, num_samples=1):
    """Generate isotropic noise on the grid.

    Sample the equivalent of white noise on a sphere and project it onto a grid using
    Driscoll and Healy, 1994 algorithm. The power spectrum is normalized to have variance 1.
    We need to assume lons = 2* lats.

    Args:
        num_lat (int): Number of latitudes in the final grid.
        num_samples (int, optional): Number of indipendent samples. Defaults to 1.

    Returns:
        grid: Numpy array with shape shape(grid) x num_samples.
    """
    power = np.ones(num_lat // 2, dtype=float) / (
        num_lat // 2
    )  # normalized to get each point with std 1
    grid = np.zeros((num_lat * 2, num_lat, num_samples))
    for i in range(num_samples):
        clm = pysh.SHCoeffs.from_random(power)
        grid[:, :, i] = clm.expand(grid="DH2", extend=False).to_array().transpose()
    return grid.astype(np.float32)


def sample_noise_level(sigma_min=0.02, sigma_max=88, rho=7):
    """Generate random sample of noise level.

    Sample a noise level according to the distribution described in the paper.
    Notice that the default values are valid only for training and need to be
    modified for sampling.

    Args:
        sigma_min (float, optional): Defaults to 0.02.
        sigma_max (int, optional): Defaults to 88.
        rho (int, optional): Defaults to 7.

    Returns:
        noise_level: single sample of noise level.
    """
    u = np.random.random()
    noise_level = (
        sigma_max ** (1 / rho) + u * (sigma_min ** (1 / rho) - sigma_max ** (1 / rho))
    ) ** rho
    return noise_level


class Preconditioner(torch.nn.Module):
    """Collection of preconditioning functions.

    These functions are described in Karras (2022), table 1.
    """

    def __init__(self, sigma_data: float = 1):
        """Initialize the preconditioning functions.

        Args:
            sigma_data (float): Karras suggests 0.5, GenCast 1. Defaults to 1.
        """
        super().__init__()
        self.sigma_data = sigma_data

    def c_skip(self, sigma):
        """Scaling factor for skip connection."""
        return self.sigma_data / (sigma**2 + self.sigma_data**2)

    def c_out(self, sigma):
        """Scaling factor for output."""
        return sigma * self.sigma_data / torch.sqrt(sigma**2 + self.sigma_data**2)

    def c_in(self, sigma):
        """Scaling factor for input."""
        return 1 / torch.sqrt(sigma**2 + self.sigma_data**2)

    def c_noise(self, sigma):
        """Scaling factor for noise level."""
        return 1 / 4 * torch.log(sigma)



"""Build the three graphs for GenCast.

The following code is a port of several components from GraphCast's original graph generation
(https://github.com/google-deepmind/graphcast) to PyG and PyTorch. The graphs are:
- g2m: grid to mesh.
- mesh: icosphere refinement.
- m2g: mesh to grid.
- khop: k-hop neighbours mesh.
"""

import numpy as np
import torch
from torch_geometric.data import Data, HeteroData
from torch_geometric.transforms import TwoHop

from graph_weather.models.gencast.graph import grid_mesh_connectivity, icosahedral_mesh, model_utils

# Some configs from graphcast:
_spatial_features_kwargs = dict(
    add_node_positions=False,
    add_node_latitude=True,
    add_node_longitude=True,
    add_relative_positions=True,
    relative_longitude_local_coordinates=True,
    relative_latitude_local_coordinates=True,
)

# radius_query_fraction_edge_length: Scalar that will be multiplied by the
#   length of the longest edge of the finest mesh to define the radius of
#   connectivity to use in the Grid2Mesh graph. Reasonable values are
#   between 0.6 and 1. 0.6 reduces the number of grid points feeding into
#   multiple mesh nodes and therefore reduces edge count and memory use, but
#   1 gives better predictions.
# mesh2grid_edge_normalization_factor: Allows explicitly controlling edge
#   normalization for mesh2grid edges. If None, defaults to max edge length.
#   This supports using pre-trained model weights with a different graph
#   structure to what it was trained on.

radius_query_fraction_edge_length = 0.6
mesh2grid_edge_normalization_factor = None


def _get_max_edge_distance(mesh):
    senders, receivers = icosahedral_mesh.faces_to_edges(mesh.faces)
    edge_distances = np.linalg.norm(mesh.vertices[senders] - mesh.vertices[receivers], axis=-1)
    return edge_distances.max()


class GraphBuilder:
    """
    Class for building GenCast's graphs.

    Attributes:
        g2m_graph (pyg.data.HeteroData): heterogeneous directed graph connecting the grid nodes
            to the mesh nodes.
        mesh_graph (pyg.data.Data): undirected graph connecting the mesh nodes.
        m2g_graph (pyg.data.HeteroData): heterogeneous directed graph connecting the mesh nodes
            to the grid nodes.
        khop_mesh_graph (pyg.data.Data): augmented version of mesh_graph in which every node is
            connected to its 2^num_hops neighbours.
        grid_nodes_dim (int): dimension of the grid nodes features.
        mesh_nodes_dim (int): dimension of the mesh nodes features.
        mesh_edges_dim (int): dimension of the mesh edges features.
        g2m_edges_dim (int): dimension of the "grid to mesh" edges features.
        m2g_edges_dim (int): dimension of the "mesh to grid" edges features.
    """

    def __init__(
        self,
        grid_lon: np.ndarray,
        grid_lat: np.ndarray,
        splits: int = 5,
        num_hops: int = 0,
        device: torch.device = torch.device("cpu"),
    ):
        """Initialize the GraphBuilder object.

        Args:
            grid_lon: 1D np.ndarray containing the list of longitudes.
            grid_lat: 1D np.ndarray containing the list of latitudes.
            splits: number of times to split the icosphere to build the mesh. Defaults to 5.
            num_hops: if num_hops=k then khop_mesh_graph will be the 2^k-neighbours version of
                the mesh. Defaults to 0.
            device: the device to which the graph will be moved.
        """

        self._spatial_features_kwargs = _spatial_features_kwargs
        self.add_edge_features_to_khop = True
        self.device = device

        # Specification of the mesh.
        _icosahedral_refinements = icosahedral_mesh.get_hierarchy_of_triangular_meshes_for_sphere(
            splits
        )
        self._mesh = _icosahedral_refinements[-1]

        # Obtain the query radius in absolute units for the unit-sphere for the
        # grid2mesh model, by rescaling the `radius_query_fraction_edge_length`.
        self._query_radius = _get_max_edge_distance(self._mesh) * radius_query_fraction_edge_length
        self._mesh2grid_edge_normalization_factor = mesh2grid_edge_normalization_factor

        self.grid_nodes_dim = None
        self.mesh_nodes_dim = None
        self.mesh_edges_dim = None
        self.g2m_edges_dim = None
        self.m2g_edges_dim = None

        # A "_init_mesh_properties":
        # This one could be initialized at init but we delay it for consistency too.
        self._num_mesh_nodes = None  # num_mesh_nodes
        self._mesh_nodes_lat = None  # [num_mesh_nodes]
        self._mesh_nodes_lon = None  # [num_mesh_nodes]

        # A "_init_grid_properties":
        self._grid_lat = None  # [num_lat_points]
        self._grid_lon = None  # [num_lon_points]
        self._num_grid_nodes = None  # num_lat_points * num_lon_points
        self._grid_nodes_lat = None  # [num_grid_nodes]
        self._grid_nodes_lon = None  # [num_grid_nodes]

        self._init_grid_properties(grid_lat, grid_lon)
        self._init_mesh_properties()
        self.g2m_graph = self._init_grid2mesh_graph()
        self.mesh_graph = self._init_mesh_graph()
        self.m2g_graph = self._init_mesh2grid_graph()

        self.num_hops = num_hops
        self.khop_mesh_graph = self._init_khop_mesh_graph()

    def _init_grid_properties(self, grid_lat: np.ndarray, grid_lon: np.ndarray):
        """Inits static properties that have to do with grid nodes."""
        self._grid_lat = grid_lat.astype(np.float32)
        self._grid_lon = grid_lon.astype(np.float32)
        # Initialized the counters.
        self._num_grid_nodes = grid_lat.shape[0] * grid_lon.shape[0]

        # Initialize lat and lon for the grid.
        grid_nodes_lon, grid_nodes_lat = np.meshgrid(grid_lon, grid_lat)
        self._grid_nodes_lon = grid_nodes_lon.reshape([-1]).astype(np.float32)
        self._grid_nodes_lat = grid_nodes_lat.reshape([-1]).astype(np.float32)

    def _init_mesh_properties(self):
        """Inits static properties that have to do with mesh nodes."""
        self._num_mesh_nodes = self._mesh.vertices.shape[0]
        mesh_phi, mesh_theta = model_utils.cartesian_to_spherical(
            self._mesh.vertices[:, 0], self._mesh.vertices[:, 1], self._mesh.vertices[:, 2]
        )
        (
            mesh_nodes_lat,
            mesh_nodes_lon,
        ) = model_utils.spherical_to_lat_lon(phi=mesh_phi, theta=mesh_theta)
        # Convert to f32 to ensure the lat/lon features aren't in f64.
        self._mesh_nodes_lat = mesh_nodes_lat.astype(np.float32)
        self._mesh_nodes_lon = mesh_nodes_lon.astype(np.float32)

    def _init_grid2mesh_graph(self):
        """Build Grid2Mesh graph."""

        # Create some edges according to distance between mesh and grid nodes.
        assert self._grid_lat is not None and self._grid_lon is not None
        (grid_indices, mesh_indices) = grid_mesh_connectivity.radius_query_indices(
            grid_latitude=self._grid_lat,
            grid_longitude=self._grid_lon,
            mesh=self._mesh,
            radius=self._query_radius,
        )

        # Edges sending info from grid to mesh.
        senders = grid_indices
        receivers = mesh_indices

        # Precompute structural node and edge features according to config options.
        # Structural features are those that depend on the fixed values of the
        # latitude and longitudes of the nodes.
        (senders_node_features, receivers_node_features, edge_features) = (
            model_utils.get_bipartite_graph_spatial_features(
                senders_node_lat=self._grid_nodes_lat,
                senders_node_lon=self._grid_nodes_lon,
                receivers_node_lat=self._mesh_nodes_lat,
                receivers_node_lon=self._mesh_nodes_lon,
                senders=senders,
                receivers=receivers,
                edge_normalization_factor=None,
                **self._spatial_features_kwargs,
            )
        )

        self.grid_nodes_dim = senders_node_features.shape[1]
        self.mesh_nodes_dim = receivers_node_features.shape[1]
        self.g2m_edges_dim = edge_features.shape[1]

        g2m_graph = HeteroData()
        g2m_graph["grid_nodes"].x = torch.tensor(
            senders_node_features, dtype=torch.float32, device=self.device
        )  # TODO: generate graph with torch or np?
        g2m_graph["mesh_nodes"].x = torch.tensor(
            receivers_node_features, dtype=torch.float32, device=self.device
        )
        g2m_graph["grid_nodes", "to", "mesh_nodes"].edge_index = torch.tensor(
            np.stack([senders, receivers]), dtype=torch.long, device=self.device
        )
        g2m_graph["grid_nodes", "to", "mesh_nodes"].edge_attr = torch.tensor(
            edge_features, dtype=torch.float32, device=self.device
        )

        return g2m_graph

    def _init_mesh_graph(self):
        """Build Mesh graph."""
        # Work simply on the mesh edges.
        senders, receivers = icosahedral_mesh.faces_to_edges(self._mesh.faces)

        # Precompute structural node and edge features according to config options.
        # Structural features are those that depend on the fixed values of the
        # latitude and longitudes of the nodes.
        assert self._mesh_nodes_lat is not None and self._mesh_nodes_lon is not None
        node_features, edge_features = model_utils.get_graph_spatial_features(
            node_lat=self._mesh_nodes_lat,
            node_lon=self._mesh_nodes_lon,
            senders=senders,
            receivers=receivers,
            **self._spatial_features_kwargs,
        )

        self.mesh_edges_dim = edge_features.shape[1]

        mesh_graph = Data(
            x=torch.tensor(node_features, dtype=torch.float32, device=self.device),
            edge_attr=torch.tensor(edge_features, dtype=torch.float32, device=self.device),
            edge_index=torch.tensor(
                np.stack([senders, receivers]), dtype=torch.long, device=self.device
            ),
        )

        return mesh_graph

    def _init_mesh2grid_graph(self):
        """Build Mesh2Grid graph."""

        # Create some edges according to how the grid nodes are contained by
        # mesh triangles.
        (grid_indices, mesh_indices) = grid_mesh_connectivity.in_mesh_triangle_indices(
            grid_latitude=self._grid_lat, grid_longitude=self._grid_lon, mesh=self._mesh
        )

        # Edges sending info from mesh to grid.
        senders = mesh_indices
        receivers = grid_indices

        # Precompute structural node and edge features according to config options.
        assert self._mesh_nodes_lat is not None and self._mesh_nodes_lon is not None
        (senders_node_features, receivers_node_features, edge_features) = (
            model_utils.get_bipartite_graph_spatial_features(
                senders_node_lat=self._mesh_nodes_lat,
                senders_node_lon=self._mesh_nodes_lon,
                receivers_node_lat=self._grid_nodes_lat,
                receivers_node_lon=self._grid_nodes_lon,
                senders=senders,
                receivers=receivers,
                edge_normalization_factor=self._mesh2grid_edge_normalization_factor,
                **self._spatial_features_kwargs,
            )
        )

        self.m2g_edges_dim = edge_features.shape[1]

        m2g_graph = HeteroData()
        m2g_graph["mesh_nodes"].x = torch.tensor(
            senders_node_features, dtype=torch.float32, device=self.device
        )
        m2g_graph["grid_nodes"].x = torch.tensor(
            receivers_node_features, dtype=torch.float32, device=self.device
        )
        m2g_graph["mesh_nodes", "to", "grid_nodes"].edge_index = torch.tensor(
            np.stack([senders, receivers]), dtype=torch.long, device=self.device
        )
        m2g_graph["mesh_nodes", "to", "grid_nodes"].edge_attr = torch.tensor(
            edge_features, dtype=torch.float32, device=self.device
        )

        return m2g_graph

    def _init_khop_mesh_graph(self):
        """Build k-hop Mesh graph."""

        transform = TwoHop()
        khop_mesh_graph = self.mesh_graph
        for _ in range(self.num_hops):
            khop_mesh_graph = transform(khop_mesh_graph)

        if self.add_edge_features_to_khop:
            senders = khop_mesh_graph.edge_index[0]
            receivers = khop_mesh_graph.edge_index[1]
            _, edge_features = model_utils.get_graph_spatial_features(
                node_lat=self._mesh_nodes_lat,
                node_lon=self._mesh_nodes_lon,
                senders=senders,
                receivers=receivers,
                **self._spatial_features_kwargs,
            )
            khop_mesh_graph.edge_attr = torch.tensor(
                edge_features, dtype=torch.float32, device=self.device
            )
        return khop_mesh_graph



"""Denoiser.

The denoiser takes as inputs the previous two timesteps, the corrupted target residual, and the
noise level, and outputs the denoised predictions. It performs the following tasks:
1. Initializes the graph, encoder, processor, and decoder.
2. Computes f_theta as the combination of encoder, processor, and decoder.
3. Preconditions f_theta on the noise levels using the parametrization from Karras et al. (2022).
"""

import einops
import numpy as np
import torch
from torch_geometric.data import Batch

from graph_weather.models.gencast.graph.graph_builder import GraphBuilder
from graph_weather.models.gencast.layers.decoder import Decoder
from graph_weather.models.gencast.layers.encoder import Encoder
from graph_weather.models.gencast.utils.noise import Preconditioner


class Denoiser(torch.nn.Module):
    """GenCast's Denoiser."""

    def __init__(
        self,
        grid_lon: np.ndarray,
        grid_lat: np.ndarray,
        input_features_dim: int,
        output_features_dim: int,
        hidden_dims: list[int] = [512, 512],
        num_blocks: int = 16,
        num_heads: int = 4,
        splits: int = 6,
        num_hops: int = 6,
        device: torch.device = torch.device("cpu"),
    ):
        """Initialize the Denoiser.

        Args:
            grid_lon (np.ndarray): array of longitudes.
            grid_lat (np.ndarray): array of latitudes.
            input_features_dim (int): dimension of the input features for a single timestep.
            output_features_dim (int): dimension of the target features.
            hidden_dims (list[int], optional): list of dimensions for the hidden layers in the MLPs
                used in GenCast. This also determines the latent dimension. Defaults to [512, 512].
            num_blocks (int, optional): number of transformer blocks in Processor. Defaults to 16.
            num_heads (int, optional): number of heads for each transformer. Defaults to 4.
            splits (int, optional): number of time to split the icosphere during graph building.
                Defaults to 6.
            num_hops (int, optional): the transformes will attention to the (2^num_hops)-neighbours
                of each node. Defaults to 6.
            device (torch.device, optional): device on which we want to build graph.
                Defaults to torch.device("cpu").
        """
        super().__init__()
        self.num_lon = len(grid_lon)
        self.num_lat = len(grid_lat)
        self.input_features_dim = input_features_dim
        self.output_features_dim = output_features_dim

        # Initialize graph
        self.graphs = GraphBuilder(
            grid_lon=grid_lon,
            grid_lat=grid_lat,
            splits=splits,
            num_hops=num_hops,
            device=device,
        )

        # Initialize Encoder
        self.encoder = Encoder(
            grid_dim=output_features_dim + 2 * input_features_dim + self.graphs.grid_nodes_dim,
            mesh_dim=self.graphs.mesh_nodes_dim,
            edge_dim=self.graphs.g2m_edges_dim,
            hidden_dims=hidden_dims,
            activation_layer=torch.nn.SiLU,
            use_layer_norm=True,
        )

        # Initialize Decoder
        self.decoder = Decoder(
            edges_dim=self.graphs.m2g_edges_dim,
            output_dim=output_features_dim,
            hidden_dims=hidden_dims,
            activation_layer=torch.nn.SiLU,
            use_layer_norm=True,
        )

        # Initialize preconditioning functions
        self.precs = Preconditioner(sigma_data=1.0)

    def _check_shapes(self, corrupted_targets, prev_inputs, noise_levels):
        batch_size = prev_inputs.shape[0]
        exp_inputs_shape = (batch_size, self.num_lon, self.num_lat, 2 * self.input_features_dim)
        exp_targets_shape = (batch_size, self.num_lon, self.num_lat, self.output_features_dim)
        exp_noise_shape = (batch_size, 1)

        if not all(
            [
                corrupted_targets.shape == exp_targets_shape,
                prev_inputs.shape == exp_inputs_shape,
                noise_levels.shape == exp_noise_shape,
            ]
        ):
            raise ValueError(
                "The shapes of the input tensors don't match with the initialization parameters: "
                f"expected {exp_inputs_shape} for prev_inputs, {exp_targets_shape} for targets and "
                f"{exp_noise_shape} for noise_levels."
            )

    def _run_encoder(self, grid_features):
        # build big graph with batch_size disconnected copies of the graph, with features [(b n) f].
        batch_size = grid_features.shape[0]
        g2m_batched = Batch.from_data_list([self.graphs.g2m_graph] * batch_size)

        # load features.
        grid_features = einops.rearrange(grid_features, "b n f -> (b n) f")
        input_grid_nodes = torch.cat([grid_features, g2m_batched["grid_nodes"].x], dim=-1).type(
            torch.float32
        )
        input_mesh_nodes = g2m_batched["mesh_nodes"].x
        input_edge_attr = g2m_batched["grid_nodes", "to", "mesh_nodes"].edge_attr
        edge_index = g2m_batched["grid_nodes", "to", "mesh_nodes"].edge_index

        # run the encoder.
        latent_grid_nodes, latent_mesh_nodes = self.encoder(
            input_grid_nodes=input_grid_nodes,
            input_mesh_nodes=input_mesh_nodes,
            input_edge_attr=input_edge_attr,
            edge_index=edge_index,
        )

        # restore nodes dimension: [b, n, f]
        latent_grid_nodes = einops.rearrange(latent_grid_nodes, "(b n) f -> b n f", b=batch_size)
        latent_mesh_nodes = einops.rearrange(latent_mesh_nodes, "(b n) f -> b n f", b=batch_size)
        return latent_grid_nodes, latent_mesh_nodes

    def _run_decoder(self, latent_mesh_nodes, latent_grid_nodes):
        # build big graph with batch_size disconnected copies of the graph, with features [(b n) f].
        batch_size = latent_mesh_nodes.shape[0]
        m2g_batched = Batch.from_data_list([self.graphs.m2g_graph] * batch_size)

        # load features.
        input_mesh_nodes = einops.rearrange(latent_mesh_nodes, "b n f -> (b n) f")
        input_grid_nodes = einops.rearrange(latent_grid_nodes, "b n f -> (b n) f")
        input_edge_attr = m2g_batched["mesh_nodes", "to", "grid_nodes"].edge_attr
        edge_index = m2g_batched["mesh_nodes", "to", "grid_nodes"].edge_index

        # run the decoder.
        output_grid_nodes = self.decoder(
            input_mesh_nodes=input_mesh_nodes,
            input_grid_nodes=input_grid_nodes,
            input_edge_attr=input_edge_attr,
            edge_index=edge_index,
        )

        # restore nodes dimension: [b, n, f]
        output_grid_nodes = einops.rearrange(output_grid_nodes, "(b n) f -> b n f", b=batch_size)
        return output_grid_nodes

    def _run_processor(self, latent_mesh_nodes, noise_levels):
        # TODO: add processor.
        return latent_mesh_nodes

    def _f_theta(self, grid_features, noise_levels):
        # run encoder, processor and decoder.
        latent_grid_nodes, latent_mesh_nodes = self._run_encoder(grid_features)
        latent_mesh_nodes = self._run_processor(latent_mesh_nodes, noise_levels)
        output_grid_nodes = self._run_decoder(latent_mesh_nodes, latent_grid_nodes)
        return output_grid_nodes

    def forward(
        self, corrupted_targets: torch.Tensor, prev_inputs: torch.Tensor, noise_levels: torch.Tensor
    ) -> torch.Tensor:
        """Compute the denoiser output.

        The denoiser is a version of the (encoder, processor, decoder)-model (called f_theta),
        preconditioned on the noise levels, as described below:

        D(Z, X, sigma) := c_skip(sigma)Z + c_out(sigma) * f_theta(c_in(sigma)Z, X, c_noise(sigma)),

        where Z is the corrupted target, X is the previous two timesteps concatenated and sigma is
        the noise level used for Z's corruption.

        Args:
            corrupted_targets (torch.Tensor): the target residuals corrupted by noise.
            prev_inputs (torch.Tensor): the previous two timesteps concatenated across the features'
                dimension.
            noise_levels (torch.Tensor): the noise level used for corruption.
        """
        # check shapes.
        self._check_shapes(corrupted_targets, prev_inputs, noise_levels)

        # flatten lon/lat dimensions.
        prev_inputs = einops.rearrange(prev_inputs, "b lon lat f -> b (lon lat) f")
        corrupted_targets = einops.rearrange(corrupted_targets, "b lon lat f -> b (lon lat) f")

        # apply preconditioning functions to target and noise.
        scaled_targets = self.precs.c_in(noise_levels)[:, :, None] * corrupted_targets
        scaled_noise_levels = self.precs.c_noise(noise_levels)

        # concatenate inputs and targets across features dimension.
        grid_features = torch.cat((scaled_targets, prev_inputs), dim=-1)

        # run the model.
        preds = self._f_theta(grid_features, scaled_noise_levels)

        # add skip connection.
        out = (
            self.precs.c_skip(noise_levels)[:, :, None] * corrupted_targets
            + self.precs.c_out(noise_levels)[:, :, None] * preds
        )

        # restore lon/lat dimensions.
        out = einops.rearrange(out, "b (lon lat) f -> b lon lat f", lon=self.num_lon)
        return out



"""Decoders to decode from the Processor graph to the original graph with updated values

In the original paper the decoder is described as

The Decoder maps back to physical data defined on a latitude/longitude grid. The underlying graph is
again bipartite, this time mapping icosahedron→lat/lon.
The inputs to the Decoder come from the Processor, plus a skip connection back to the original
state of the 78 atmospheric variables onthe latitude/longitude grid.
The output of the Decoder is the predicted 6-hour change in the 78 atmospheric variables,
which is then added to the initial state to produce the new state. We found 6 hours to be a good
balance between shorter time steps (simpler dynamics to model but more iterations required during
rollout) and longer time steps (fewer iterations required during rollout but modeling
more complex dynamics)

"""

import einops
import h3
import numpy as np
import torch
from torch_geometric.data import Data

from graph_weather.models.layers.graph_net_block import MLP, GraphProcessor


class AssimilatorDecoder(torch.nn.Module):
    """Assimilator graph module"""

    def __init__(
        self,
        lat_lons: list,
        resolution: int = 2,
        input_dim: int = 256,
        output_dim: int = 78,
        output_edge_dim: int = 256,
        hidden_dim_processor_node: int = 256,
        hidden_dim_processor_edge: int = 256,
        hidden_layers_processor_node: int = 2,
        hidden_layers_processor_edge: int = 2,
        mlp_norm_type: str = "LayerNorm",
        hidden_dim_decoder: int = 128,
        hidden_layers_decoder: int = 2,
        use_checkpointing: bool = False,
    ):
        """
        Decoder from latent graph to lat/lon graph for assimilation of observation

        Args:
            lat_lons: List of (lat,lon) points
            resolution: H3 resolution level
            input_dim: Input node dimension
            output_dim: Output node dimension
            output_edge_dim: Edge dimension
            hidden_dim_processor_node: Hidden dimension of the node processors
            hidden_dim_processor_edge: Hidden dimension of the edge processors
            hidden_layers_processor_node: Number of hidden layers in the node processors
            hidden_layers_processor_edge: Number of hidden layers in the edge processors
            hidden_dim_decoder:Number of hidden dimensions in the decoder
            hidden_layers_decoder: Number of layers in the decoder
            mlp_norm_type: Type of norm for the MLPs
                one of 'LayerNorm', 'GraphNorm', 'InstanceNorm', 'BatchNorm', 'MessageNorm', or None
            use_checkpointing: Whether to use gradient checkpointing to reduce model size
        """
        super().__init__()
        self.use_checkpointing = use_checkpointing
        self.num_latlons = len(lat_lons)
        self.base_h3_grid = sorted(list(h3.uncompact(h3.get_res0_indexes(), resolution)))
        self.num_h3 = len(self.base_h3_grid)
        self.h3_grid = [h3.geo_to_h3(lat, lon, resolution) for lat, lon in lat_lons]
        self.h3_to_index = {}
        h_index = len(self.base_h3_grid)
        for h in self.base_h3_grid:
            if h not in self.h3_to_index:
                h_index -= 1
                self.h3_to_index[h] = h_index
        self.h3_mapping = {}
        for h, value in enumerate(self.h3_grid):
            self.h3_mapping[h + self.num_h3] = value

        # Build the default graph
        # Extra starting ones for appending to inputs, could 'learn' good starting points
        self.latlon_nodes = torch.zeros((len(lat_lons), input_dim), dtype=torch.float)
        # Get connections between lat nodes and h3 nodes TODO Paper makes it seem like the 3
        #  closest iso points map to the lat/lon point Do kring 1 around current h3 cell,
        #  and calculate distance between all those points and the lat/lon one, choosing the
        #  nearest N (3) For a bit simpler, just include them all with their distances
        edge_sources = []
        edge_targets = []
        self.h3_to_lat_distances = []
        for node_index, h_node in enumerate(self.h3_grid):
            # Get h3 index
            h_points = h3.k_ring(self.h3_mapping[node_index + self.num_h3], 1)
            for h in h_points:
                distance = h3.point_dist(lat_lons[node_index], h3.h3_to_geo(h), unit="rads")
                self.h3_to_lat_distances.append([np.sin(distance), np.cos(distance)])
                edge_sources.append(self.h3_to_index[h])
                edge_targets.append(node_index + self.num_h3)
        edge_index = torch.tensor([edge_sources, edge_targets], dtype=torch.long)
        self.h3_to_lat_distances = torch.tensor(self.h3_to_lat_distances, dtype=torch.float)

        # Use normal graph as its a bit simpler
        self.graph = Data(edge_index=edge_index, edge_attr=self.h3_to_lat_distances)

        self.edge_encoder = MLP(
            2, output_edge_dim, hidden_dim_processor_edge, 2, mlp_norm_type, self.use_checkpointing
        )
        self.graph_processor = GraphProcessor(
            mp_iterations=1,
            in_dim_node=input_dim,
            in_dim_edge=output_edge_dim,
            hidden_dim_node=hidden_dim_processor_node,
            hidden_dim_edge=hidden_dim_processor_edge,
            hidden_layers_node=hidden_layers_processor_node,
            hidden_layers_edge=hidden_layers_processor_edge,
            norm_type=mlp_norm_type,
        )
        self.node_decoder = MLP(
            input_dim,
            output_dim,
            hidden_dim_decoder,
            hidden_layers_decoder,
            None,
            self.use_checkpointing,
        )

    def forward(self, processor_features: torch.Tensor, batch_size: int) -> torch.Tensor:
        """
        Adds features to the encoding graph

        Args:
            processor_features: Processed features in shape [B*Nodes, Features]
            batch_size: Batch size

        Returns:
            Updated features for model
        """
        self.graph = self.graph.to(processor_features.device)
        edge_attr = self.edge_encoder(self.graph.edge_attr)  # Update attributes based on distance
        edge_attr = einops.repeat(edge_attr, "e f -> (repeat e) f", repeat=batch_size)

        edge_index = torch.cat(
            [
                self.graph.edge_index + i * torch.max(self.graph.edge_index) + i
                for i in range(batch_size)
            ],
            dim=1,
        )

        # Readd nodes to match graph node number
        self.latlon_nodes = self.latlon_nodes.to(processor_features.device)
        features = einops.rearrange(processor_features, "(b n) f -> b n f", b=batch_size)
        features = torch.cat(
            [features, einops.repeat(self.latlon_nodes, "n f -> b n f", b=batch_size)], dim=1
        )
        features = einops.rearrange(features, "b n f -> (b n) f")

        out, _ = self.graph_processor(features, edge_index, edge_attr)  # Message Passing
        # Remove the h3 nodes now, only want the latlon ones
        out = self.node_decoder(out)  # Decode to 78 from 256
        out = einops.rearrange(out, "(b n) f -> b n f", b=batch_size)
        test, out = torch.split(out, [self.num_h3, self.num_latlons], dim=1)
        return out
